#
# Copyright (c) 2022 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Fri Jan  7 2022
#
"""Communication with genomixd servers."""

import json
import re
import socket
import urllib.parse
import warnings

from . import event

class Client:
  """Minimal http client for genomixd server."""

  # --- __init__ -----------------------------------------------------------
  #
  def __init__(self, host):
    """Establish a new connection to the genomixd server at 'host'."""

    addr = re.split(r':([^]:]+)$', host)
    port = addr[1] if len(addr) > 1 else 8080

    self.__rqst = []
    self.__buffer = b''
    self.__eof = False
    self.__sock = socket.create_connection((addr[0], port))

    # non-blocking required only for async sock_recv
    self.__sock.setblocking(False)

    # default watcher
    event.mux.watch(self.__sock, self.__data)


  # --- close --------------------------------------------------------------
  #
  def close(self):
    """Close connection."""
    event.mux.watch(self.__sock, None)
    self.__sock.close()


  # --- get ----------------------------------------------------------------
  #
  # http state machine
  __DONE = 0
  __STATUS = 1
  __HEADERS = 2
  __BODY = 3

  def get(self, uri, args = {}, body = None):
    """Send a http get request and return the result.

    Other genomix events are processed in the meantime.
    """

    # build request string
    r = 'get ' + uri
    if args:
      r += '?' + urllib.parse.urlencode(args)
    r += ' HTTP/1.0\n'

    if body is not None:
      r += 'Content-Type: application/x-www-form-urlencoded\n'
      r += 'Content-Length: %d\n' % len(body)
    r += '\n'

    if body is not None:
      r += body

    # send
    try:
      self.__sock.sendall(r.encode())
    except OSError:
      return {'state': self.__DONE,
              'status': 500, 'body': 'connection closed'}

    # push new state
    rqst = { 'state': self.__STATUS }
    self.__rqst += [rqst]

    try:
      # wait for completion. This might recurse, as a consquence of processing
      # callbacks in the meantime.
      while rqst['state'] != self.__DONE:
        event.mux.wait()

    finally:
      # pop state and clean watcher when the last pending request is done
      self.__rqst.pop()

    # decode the reply. All replies are JSON messages, but genomixd sends
    # invalid JSON strings (missing quotes) when the reply is a bare string.
    # Fixing genomix would break all clients, so deal with this here.
    try:
      rqst['body'] = json.loads(rqst['body'].decode())
    except json.JSONDecodeError:
      rqst['body'] = rqst['body'].decode()

    return rqst


  # --- events -------------------------------------------------------------
  #
  def events(self, data = None):
    """Iterator on server side events. Returns dictionaries of events."""

    # fill internal buffer
    self.__feed(data)

    e = {}
    for l in self.__fgets():
      if l is None: break

      if not l:
        if 'data' in e:
          data = '\n'.join(e['data'])
          try:
            e['data'] = json.loads(data)
          except json.JSONDecodeError:
            e['data'] = data

        yield e
        e = {}

      k, _, v = l.partition(':')
      e.setdefault(k, []).append(v.lstrip(' '))


  # --- watch --------------------------------------------------------------
  #
  def watch(self, cb):
    """Attach a callback invoked when readable data exists."""
    event.mux.watch(self.__sock, None)
    if cb is not None: event.mux.watch(self.__sock, cb)


  # --- __data -------------------------------------------------------------
  #
  def __data(self, data = None):
    """Process available data."""

    # fill internal buffer
    self.__feed(data)

    # there should be at least one pending request
    if not self.__rqst: return

    # iterate over requests in FIFO order until all data is exhausted
    for rqst in self.__rqst:
      # line iterator on buffered data
      lines = self.__fgets()

      try:
        if rqst['state'] == self.__STATUS:
          # process one status line (the loop prevents StopIteration)
          for l in lines:
            if l is None: raise EOFError

            rqst.update({
              'state': self.__HEADERS,
              'status': int(l.split()[1]), 'headers': {}, 'body': b''
            })
            break
          else: return

        if rqst['state'] == self.__HEADERS:
          # process headers until empty line
          for l in lines:
            if l is None: raise EOFError
            if not l:
              rqst['state'] = self.__BODY
              break

            k, _, v = l.partition(':')
            rqst['headers'][k.lower()] = v.strip()
          else: return

        if rqst['state'] == self.__BODY:
          # process body
          l = int(rqst['headers'].get('content-length', 0))
          if l:
            b = self.__read(l)
            if b is None: raise EOFError
            rqst['body'] += b

          if len(rqst['body']) >= l:
            rqst['state'] = self.__DONE
          else: return

      except EOFError:
        rqst.update({'state': self.__DONE,
                     'status': 500, 'body': b'broken connection'})
      except Exception:
        rqst.update({'state': self.__DONE,
                     'status': 500, 'body': b'not a genomixd server'})


  # --- __fgets ------------------------------------------------------------
  #
  def __fgets(self):
    """Iterator on lines, with \n removed. Returns None on eof."""

    # iterate over complete lines
    while True:
      line, sep, rest = self.__buffer.partition(b'\n')
      if not sep: break

      self.__buffer = rest
      yield line.decode()

    # return last incomplete line + None if eof
    if self.__eof:
      if line: yield line.decode()
      yield None


  # --- __read -------------------------------------------------------------
  #
  def __read(self, size):
    """Return at most 'size' bytes from internal buffer, or None on eof."""
    s = len(self.__buffer)
    if s >= size:
      b = self.__buffer[0:size]
      self.__buffer = self.__buffer[size:]
      return b

    if self.__buffer:
      b = self.__buffer
      self.__buffer = b''
      return b

    return None if self.__eof else b''


  # --- __feed -------------------------------------------------------------
  #
  def __feed(self, data):
    """Append new data to internal buffer."""
    if data:
      # perform \r\n translation
      self.__buffer += data.replace(b'\r\n', b'\n')
    elif data == b'':
      self.__eof = True
